# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from twisted.internet.defer import succeed, Deferred
from twisted.internet.protocol import Protocol
from twisted.internet.error import ConnectError, ConnectionLost, ConnectionDone

from twisted.python.failure import Failure

from silverberg.thrift_client import (
    OnDemandThriftClient,
    ClientDisconnecting,
    ClientConnecting,
    _LossNotifyingWrapperProtocol,
    _ThriftClientFactory
)

from silverberg.cassandra.Cassandra import Client
from silverberg.test.util import BaseTestCase


class _TestConnectError(ConnectError):
    pass


class _TestConnectionDone(ConnectionDone):
    pass


class _TestConnectionLost(ConnectionLost):
    pass


class OnDemandThriftClientTests(BaseTestCase):
    def setUp(self):
        self.factory_patcher = mock.patch('silverberg.thrift_client._ThriftClientFactory')

        self.factory = self.factory_patcher.start()

        def _create_factory(client_class, connection_lost):
            self.connection_lost = connection_lost

        self.factory.side_effect = _create_factory

        self.addCleanup(self.factory_patcher.stop)

        self.endpoint = mock.Mock()
        self.twisted_transport = mock.Mock()
        self.thrift_cassandra_client = mock.Mock(Client)

        def _connect(factory):
            self.connect_d = Deferred()
            wrapper = mock.Mock()
            wrapper.transport = self.twisted_transport
            wrapper.wrapped.client = self.thrift_cassandra_client
            return self.connect_d.addCallback(lambda _: wrapper)

        self.endpoint.connect.side_effect = _connect

        self.client = OnDemandThriftClient(self.endpoint, mock.Mock())

    def test_initial_connect(self):
        d = self.client.connection()

        self.connect_d.callback(None)

        self.assertEqual(self.assertFired(d), self.thrift_cassandra_client)

    def test_initial_handshake_non_deferred(self):
        def _mock_handshake(client):
            return None

        m = mock.MagicMock(side_effect=_mock_handshake)

        d = self.client.connection(m)

        self.connect_d.callback(None)

        self.assertEqual(self.assertFired(d), self.thrift_cassandra_client)
        m.assert_called_once()

    def test_initial_handshake_deferred(self):
        def _mock_handshake_d(client):
            return succeed(None)

        m = mock.MagicMock(side_effect=_mock_handshake_d)

        d = self.client.connection(m)
        self.connect_d.callback(None)

        self.assertEqual(self.assertFired(d), self.thrift_cassandra_client)
        m.assert_called_once()

    def test_connected(self):
        d1 = self.client.connection()

        self.connect_d.callback(None)

        d2 = self.client.connection()

        self.assertNotIdentical(d1, d2)

        self.assertEqual(self.assertFired(d2), self.thrift_cassandra_client)

    def test_connect_while_connecting(self):
        d1 = self.client.connection()

        d2 = self.client.connection()

        self.assertNotIdentical(d1, d2)

        self.connect_d.callback(None)

        self.assertEqual(self.assertFired(d1), self.thrift_cassandra_client)
        self.assertEqual(self.assertFired(d2), self.thrift_cassandra_client)

    def test_connect_failed(self):
        d = self.client.connection()

        self.connect_d.errback(_TestConnectError())

        self.assertFailed(d, _TestConnectError)

    def test_connection_lost_cleanly(self):
        d = self.client.connection()

        self.connect_d.callback(None)
        self.assertFired(d)

        self.connection_lost(Failure(_TestConnectionDone()))

    def test_connection_lost_uncleanly(self):
        d = self.client.connection()

        self.connect_d.callback(None)

        self.assertFired(d)

        reason = Failure(_TestConnectionLost())
        self.connection_lost(reason)

        self.assertEqual(
            self.flushLoggedErrors(),
            [reason]
        )

    def test_disconnect(self):
        d1 = self.client.connection()

        self.connect_d.callback(None)

        self.assertFired(d1)

        d2 = self.client.disconnect()

        self.twisted_transport.loseConnection.assert_called_once()

        self.connection_lost(Failure(_TestConnectionDone()))

        self.assertFired(d2)

    def test_connect_while_disconnecting(self):
        d1 = self.client.connection()

        self.connect_d.callback(None)

        self.assertFired(d1)

        d2 = self.client.disconnect()

        d3 = self.client.connection()

        self.connection_lost(Failure(_TestConnectionDone()))

        self.assertFired(d2)

        self.assertFailed(d3, ClientDisconnecting)

    def test_disconnect_while_connecting(self):
        d1 = self.client.connection()
        d2 = self.client.disconnect()

        self.assertFailed(d2, ClientConnecting)
        self.assertEqual(
            len(self.twisted_transport.loseConnection.mock_calls), 0,
            "loseConnection should not be called since not connected")

        self.connect_d.callback(None)

        self.assertFired(d1)

    def test_disconnect_while_disconnecting(self):
        self.client.connection()
        self.connect_d.callback(None)

        d1 = self.client.disconnect()
        self.twisted_transport.reset_mock()
        d2 = self.client.disconnect()

        self.connection_lost(Failure(_TestConnectionDone()))
        self.assertEqual(
            len(self.twisted_transport.loseConnection.mock_calls), 0,
            "loseConnection should not be called since not connected")

        self.assertFired(d1)
        self.assertFired(d2)

    def test_disconnect_while_not_connected(self):
        d1 = self.client.disconnect()
        self.assertIdentical(self.assertFired(d1), None)
        self.assertEqual(
            len(self.twisted_transport.loseConnection.mock_calls), 0,
            "loseConnection should not be called since not connected")


class LossNotifyingWrapperProtocolTests(BaseTestCase):
    def setUp(self):
        self.wrapped = mock.Mock(Protocol)

        self.connection_lost = mock.Mock()

        self.wrapper = _LossNotifyingWrapperProtocol(
            self.wrapped,
            self.connection_lost
        )

    def test_makeConnection(self):
        transport = mock.Mock()
        self.wrapper.makeConnection(transport)
        self.wrapped.makeConnection.assert_called_once_with(transport)

    def test_dataReceived(self):
        self.wrapper.dataReceived('foo')
        self.wrapped.dataReceived.assert_called_once_with('foo')

    def test_connectionLost(self):
        reason = Failure(_TestConnectionDone())
        self.wrapper.connectionLost(reason)
        self.wrapped.connectionLost.assert_called_once_with(reason)


class ThriftClientFactoryTests(BaseTestCase):
    def setUp(self):
        self.client_class = mock.Mock()
        self.connection_lost = mock.Mock()
        self.factory = _ThriftClientFactory(self.client_class, self.connection_lost)

    def test_buildProtocol(self):
        protocol = self.factory.buildProtocol(mock.Mock())

        self.assertIsInstance(protocol, _LossNotifyingWrapperProtocol)



# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from twisted.internet.defer import DeferredList

from silverberg.client import CQLClient

from twisted.internet.error import ConnectError


class RoundRobinCassandraCluster(object):
    """
    Maintain several :py:class:`silverberg.client.CQLClient` instances
    connected `seed_endpoints` using `keyspace`.  Each time :py:func:`execute`
    is called a client will be selected in a round-robin fashion.

    :param seed_endpoints: A list of `IStreamClientEndpoint` providers to maintain
        client connections to.
    :param str keyspace: The cassandra keyspace to use.

    :param str user: Optional username.
    :param str password: Optional password.
    """

    def __init__(self, seed_endpoints, keyspace, user=None, password=None):
        self._seed_clients = [
            CQLClient(endpoint, keyspace, user, password)
            for endpoint in seed_endpoints
        ]
        self._client_idx = 0

    def execute(self, *args, **kwargs):
        """
        See :py:func:`silverberg.client.CQLClient.execute`
        """
        num_clients = len(self._seed_clients)
        start_client = (self._client_idx + 1) % num_clients

        def _client_error(failure, client_i):
            failure.trap(ConnectError)
            client_i = (client_i + 1) % num_clients
            if client_i == start_client:
                return failure
            else:
                return _try_execute(client_i)

        def _try_execute(client_i):
            self._client_idx = client_i
            d = self._seed_clients[client_i].execute(*args, **kwargs)
            return d.addErrback(_client_error, client_i)

        return _try_execute(start_client)

    def disconnect(self):
        """
        Disconnect from the cassandra cluster.  Cassandara and Silverberg do
        not require the connection to be closed before exiting.  However, this
        method may be useful if resources are constrained, or for testing
        purposes.

        :return: a :class:`DeferredList` that fires with a list of None's when every client
        has disconnected.
        """
        return DeferredList([client.disconnect() for client in self._seed_clients])



# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Locking recipe for Cassandra

"""

import uuid

from twisted.internet import defer, task

from silverberg.client import ConsistencyLevel
from silverberg.cassandra.ttypes import InvalidRequestException


class BusyLockError(Exception):
    def __init__(self, lock_table, lock_id):
        super(BusyLockError, self).__init__(
            "Unable to acquire lock {id} on {table}".format(id=lock_id,
                                                            table=lock_table))


class NoLockClaimsError(Exception):
    def __init__(self, lock_table, lock_id):
        super(NoLockClaimsError, self).__init__(
            "No claims were found for lock {id} on {table}".format(
                id=lock_id,
                table=lock_table))


class BasicLock(object):
    """A locking mechanism for Cassandra.

    Based on the lock implementation from Netflix's astyanax, the lock recipe
    is a write, read, write operation. A record is written to the specified
    Cassandra database table with a timeuuid, and then the table is read for
    the given lock, ordered by timeuuid. If the first row is not ours, the
    lock was not acquired, so a write to remove the lock is made.

    :param client: A Cassandra CQL client
    :type client: silverberg.client.CQLClient

    :param lock_table: A table/columnfamily table name for holding locks.
    :type lock_table: str

    :param lock_id: A unique identifier for the lock.
    :type lock_id: str

    :param ttl: A TTL for the lock.
    :type ttl: int

    :param max_retry: A number of times to retry acquisition of the lock.
    :type max_retry: int

    :param retry_wait: A number of seconds to wait before retrying acquisition.
    :type retry_wait: int

    :param reactor: A twisted clock.
    :type reactor: twisted.internet.interfaces.IReactorTime
    """

    def __init__(self, client, lock_table, lock_id, ttl=300, max_retry=0,
                 retry_wait=10, reactor=None, log=None):
        self._client = client
        self._lock_table = lock_table
        self._lock_id = lock_id
        self._claim_id = uuid.uuid1()
        self._ttl = ttl
        self._max_retry = max_retry
        self._retry_wait = retry_wait
        if reactor is None:
            from twisted.internet import reactor
        self._reactor = reactor

        self._log = log
        self._log_kwargs = dict(lock_id=self._lock_id, claim_id=self._claim_id)
        self._lock_acquired_seconds = 0
        self._acquire_start_seconds = 0

    def _read_lock(self, ignored):
        query = 'SELECT * FROM {cf} WHERE "lockId"=:lockId ORDER BY "claimId";'
        return self._client.execute(query.format(cf=self._lock_table),
                                    {'lockId': self._lock_id}, ConsistencyLevel.QUORUM)

    def _verify_lock(self, response):
        if len(response) == 0:
            return self.release().addCallback(lambda _: defer.fail(
                NoLockClaimsError(self._lock_table, self._lock_id)))

        if response[0]['claimId'] == self._claim_id:
            return defer.succeed(True)
        else:
            return self.release().addCallback(lambda _: defer.fail(
                BusyLockError(self._lock_table, self._lock_id)))

    def _write_lock(self):
        query = 'INSERT INTO {cf} ("lockId","claimId") VALUES (:lockId,:claimId) USING TTL {ttl};'
        return self._client.execute(query.format(cf=self._lock_table, ttl=self._ttl),
                                    {'lockId': self._lock_id, 'claimId': self._claim_id},
                                    ConsistencyLevel.QUORUM)

    @staticmethod
    def ensure_schema(client, table_name):
        """
        Create the table/columnfamily if it doesn't already exist.

        :param client: A Cassandra CQL client
        :type client: silverberg.client.CQLClient

        :param lock_table: A table/columnfamily table name for holding locks.
        :type lock_table: str
        """
        query = ''.join([
            'CREATE TABLE {cf} ',
            '("lockId" ascii, "claimId" timeuuid, PRIMARY KEY("lockId", "claimId"));'])

        def errback(failure):
            failure.trap(InvalidRequestException)

        return client.execute(query.format(cf=table_name),
                              {}, ConsistencyLevel.QUORUM).addErrback(errback)

    @staticmethod
    def drop_schema(client, table_name):
        """
        Delete the table/columnfamily.

        :param client: A Cassandra CQL client
        :type client: silverberg.client.CQLClient

        :param lock_table: A table/columnfamily table name for holding locks.
        :type lock_table: str
        """
        query = 'DROP TABLE {cf}'
        return client.execute(query.format(cf=table_name),
                              {}, ConsistencyLevel.QUORUM)

    def release(self):
        """
        Release the lock.
        """
        query = 'DELETE FROM {cf} WHERE "lockId"=:lockId AND "claimId"=:claimId;'
        d = self._client.execute(query.format(cf=self._lock_table),
                                 {'lockId': self._lock_id, 'claimId': self._claim_id},
                                 ConsistencyLevel.QUORUM)

        def _log_release_time(result):
            if self._log:
                seconds = self._reactor.seconds() - self._lock_acquired_seconds
                self._log.msg('Released lock. Was held for {} seconds'.format(seconds),
                              lock_held_time=seconds, **self._log_kwargs)
            return result

        return d.addBoth(_log_release_time)

    def acquire(self):
        """
        Acquire the lock.

        If the lock can't be acquired immediately, retry a specified number of
        times, with a specified wait time.
        """
        retries = [0]
        self._acquire_start_seconds = self._reactor.seconds()

        def log_lock_acquired(result):
            self._lock_acquired_seconds = self._reactor.seconds()
            seconds = self._lock_acquired_seconds - self._acquire_start_seconds
            self._log.msg('Acquired lock in {} seconds'.format(seconds),
                          lock_acquire_time=seconds, **self._log_kwargs)
            return result

        def acquire_lock():
            d = self._write_lock()
            d.addCallback(self._read_lock)
            d.addCallback(self._verify_lock)
            if self._log:
                d.addCallback(log_lock_acquired)
            d.addErrback(lock_not_acquired)
            return d

        def lock_not_acquired(failure):
            failure.trap(BusyLockError, NoLockClaimsError)
            retries[0] += 1
            if retries[0] <= self._max_retry:
                return task.deferLater(self._reactor, self._retry_wait, acquire_lock)
            else:
                return failure

        def log_lock_acquire_failure(failure):
            if self._log:
                seconds = self._reactor.seconds() - self._acquire_start_seconds
                self._log.msg(
                    'Could not acquire lock in {} seconds due to {}'.format(seconds, failure),
                    lock_acquire_fail_time=seconds, reason=failure, **self._log_kwargs)
            return failure

        return acquire_lock().addErrback(log_lock_acquire_failure)


def with_lock(lock, func, *args, **kwargs):
    """A 'context manager' for performing operations requiring a lock.

    :param lock: A BasicLock instance
    :type lock: silverberg.lock.BasicLock

    :param func: A callable to execute while the lock is held.
    :type func: function
    """
    d = lock.acquire()

    def release_lock(result):
        deferred = lock.release()
        return deferred.addCallback(lambda x: result)

    def lock_acquired(lock):
        return defer.maybeDeferred(func, *args, **kwargs).addBoth(release_lock)

    d.addCallback(lock_acquired)
    return d



# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from silverberg.test.util import BaseTestCase

from silverberg.client import CQLClient

from silverberg.cluster import RoundRobinCassandraCluster

from twisted.internet.error import ConnectionRefusedError, NoRouteError
from twisted.internet import defer


class RoundRobinCassandraClusterTests(BaseTestCase):
    """
    Tests for mod:`silverberg.cluster`
    """

    def setUp(self):
        """
        Mock CQLClient
        """
        self.cql_client_patcher = mock.patch("silverberg.cluster.CQLClient")
        self.CQLClient = self.cql_client_patcher.start()
        self.addCleanup(self.cql_client_patcher.stop)

        self.clients = []

        def _CQLClient(*args, **kwargs):
            c = mock.Mock(CQLClient)
            self.clients.append(c)
            c.execute.return_value = defer.succeed('exec_ret{}'.format(len(self.clients)))
            return c

        self.CQLClient.side_effect = _CQLClient

    def test_round_robin_execute(self):
        """
        seed_clients are chosen in round robin manner
        """
        cluster = RoundRobinCassandraCluster(['one', 'two', 'three'], 'keyspace')

        for client, arg in [(1, 'foo'), (2, 'bar'), (0, 'baz'), (1, 'bax')]:
            result = cluster.execute(arg)
            self.clients[client].execute.assert_called_with(arg)
            self.assertEqual(self.clients[client].execute.return_value, result)

    def test_one_node_down(self):
        """
        If a cass node is down, it tries the next node in cluster
        """
        cluster = RoundRobinCassandraCluster(['one', 'two', 'three'], 'keyspace')
        self.clients[1].execute.return_value = defer.fail(ConnectionRefusedError())
        result = cluster.execute(2, 3)
        self.assertEqual(self.successResultOf(result), 'exec_ret3')
        self.clients[1].execute.assert_called_once_with(2, 3)
        self.clients[2].execute.assert_called_once_with(2, 3)
        self.assertFalse(self.clients[0].execute.called)

    def test_two_nodes_down(self):
        """
        If 2 cass nodes are down, it tries the next node in cluster
        """
        cluster = RoundRobinCassandraCluster(['one', 'two', 'three'], 'keyspace')
        self.clients[1].execute.return_value = defer.fail(ConnectionRefusedError())
        self.clients[2].execute.return_value = defer.fail(NoRouteError())
        result = cluster.execute(2, 3)
        self.clients[1].execute.assert_called_once_with(2, 3)
        self.clients[2].execute.assert_called_once_with(2, 3)
        self.clients[0].execute.assert_called_once_with(2, 3)
        self.assertEqual(self.successResultOf(result), 'exec_ret1')

    def test_other_error_propogated(self):
        """
        Any error other than subclass of ``ConnectError`` is propgoated
        """
        cluster = RoundRobinCassandraCluster(['one', 'two', 'three'], 'keyspace')
        rand_err = ValueError('random err')
        self.clients[1].execute.return_value = defer.fail(rand_err)
        result = cluster.execute(2, 3)
        self.assertEqual(self.failureResultOf(result).value, rand_err)
        self.assertEqual(self.clients[2].execute.called, False)
        self.assertEqual(self.clients[0].execute.called, False)

    def test_other_error_propogated_on_node_down(self):
        """
        If first node gives ``ConnectError`` then second node is tried if it gives
        error other than subclass of ``ConnectError`` it is propogated
        """
        cluster = RoundRobinCassandraCluster(['one', 'two', 'three'], 'keyspace')
        self.clients[1].execute.return_value = defer.fail(NoRouteError())
        rand_err = ValueError('random err')
        self.clients[2].execute.return_value = defer.fail(rand_err)
        result = cluster.execute(2, 3)
        self.assertEqual(self.failureResultOf(result).value, rand_err)
        self.assertEqual(self.clients[0].execute.called, False)

    def test_all_nodes_down(self):
        """
        If all cass nodes are down, it gives up eventually by raising the
        connection error exception
        """
        cluster = RoundRobinCassandraCluster(['one', 'two', 'three'], 'keyspace')
        err = ConnectionRefusedError()
        for i in range(3):
            self.clients[i].execute.side_effect = lambda *_: defer.fail(err)

        result = cluster.execute(2, 3)

        self.assertEqual(self.failureResultOf(result).value, err)
        for i in range(3):
            self.clients[i].execute.assert_called_once_with(2, 3)

    def test_multiple_clients(self):
        """
        When multiple clients execute simultaneously, it does not skip nodes. It basically ensures
        that http://bit.ly/1csEIRR situation does not happen
        """
        cluster = RoundRobinCassandraCluster(['one', 'two', 'three'], 'keyspace')

        def _execute(*args):
            cluster.execute(2, 3)
            return defer.fail(NoRouteError())
        self.clients[1].execute.side_effect = _execute

        result = cluster.execute(2, 3)

        self.assertEqual(self.successResultOf(result), 'exec_ret3')
        self.assertEqual(self.clients[2].execute.call_args_list, [mock.call(2, 3)] * 2)
        self.clients[1].execute.assert_called_once_with(2, 3)

    def test_disconnect(self):
        cluster = RoundRobinCassandraCluster(['one', 'two', 'three'], 'keyspace')
        cluster.disconnect()

        for client in self.clients:
            client.disconnect.assert_called_with()
